import dataclasses
import time
from socket import AF_INET6, inet_ntop
from typing import List

from ifaddr import Adapter, get_adapters
from zeroconf import ServiceBrowser, ServiceListener, Zeroconf
from zeroconf.const import _TYPE_AAAA

DEFAULT_BONJOUR_TIMEOUT = 1


class RemotedListener(ServiceListener):
    def __init__(self, adapter: Adapter):
        super().__init__()
        self.adapter = adapter
        self.addresses: List[str] = []

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        if name == 'ncm._remoted._tcp.local.':
            service_info = zc.get_service_info(type_, name)
            entries_with_name = zc.cache.async_entries_with_name(service_info.server)
            for entry in entries_with_name:
                if entry.type == _TYPE_AAAA:
                    self.addresses.append(inet_ntop(AF_INET6, entry.address) + '%' + self.adapter.nice_name)

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        pass

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        pass


@dataclasses.dataclass
class BonjourQuery:
    zc: Zeroconf
    service_browser: ServiceBrowser
    listener: RemotedListener


def query_bonjour(adapter: Adapter) -> BonjourQuery:
    zc = Zeroconf(interfaces=[adapter.ips[0].ip[0]])
    listener = RemotedListener(adapter)
    service_browser = ServiceBrowser(zc, '_remoted._tcp.local.', listener)
    return BonjourQuery(zc, service_browser, listener)


def get_remoted_addresses(timeout: int = DEFAULT_BONJOUR_TIMEOUT) -> List[str]:
    adapters = [adapter for adapter in get_adapters() if adapter.ips[0].is_IPv6]
    bonjour_queries = [query_bonjour(adapter) for adapter in adapters]
    time.sleep(timeout)
    addresses = []
    for bonjour_query in bonjour_queries:
        addresses += bonjour_query.listener.addresses
        bonjour_query.service_browser.cancel()
        bonjour_query.zc.close()
    return addresses



from typing import Union

from packaging.version import Version

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.remote_server import RemoteServer


class DvtSecureSocketProxyService(RemoteServer):
    SERVICE_NAME = 'com.apple.instruments.remoteserver.DVTSecureSocketProxy'
    OLD_SERVICE_NAME = 'com.apple.instruments.remoteserver'
    RSD_SERVICE_NAME = 'com.apple.instruments.dtservicehub'

    def __init__(self, lockdown: Union[RemoteServiceDiscoveryService, LockdownClient]):
        if isinstance(lockdown, RemoteServiceDiscoveryService):
            service_name = self.RSD_SERVICE_NAME
            remove_ssl_context = False
        elif Version(lockdown.product_version) >= Version('14.0'):
            service_name = self.SERVICE_NAME
            remove_ssl_context = False
        else:
            service_name = self.OLD_SERVICE_NAME
            remove_ssl_context = True

        super().__init__(lockdown, service_name, remove_ssl_context=remove_ssl_context)



import contextlib
import platform
from typing import Generator

import psutil

from pymobiledevice3.exceptions import AccessDeniedError

REMOTED_PATH = '/usr/libexec/remoted'


def get_remoted_process() -> psutil.Process:
    for process in psutil.process_iter():
        if process.pid == 0:
            # skip kernel task
            continue
        try:
            if process.exe() == REMOTED_PATH:
                return process
        except (psutil.ZombieProcess, psutil.NoSuchProcess):
            continue


def stop_remoted_if_required() -> None:
    if platform.system() != 'Darwin':
        # only Darwin systems require it
        return

    remoted = get_remoted_process()
    if remoted is None:
        return
    if remoted.status() == 'stopped':
        # process already stopped, we don't need to do anything
        return

    try:
        remoted.suspend()
    except psutil.AccessDenied:
        raise AccessDeniedError()


def resume_remoted_if_required() -> None:
    if platform.system() != 'Darwin':
        # only Darwin systems require it
        return

    remoted = get_remoted_process()
    if remoted is None:
        return
    if remoted.status() == 'running':
        # process already running, we don't need to do anything
        return

    try:
        remoted.resume()
    except psutil.AccessDenied:
        raise AccessDeniedError()


@contextlib.contextmanager
def stop_remoted() -> Generator[None, None, None]:
    stop_remoted_if_required()
    try:
        yield
    finally:
        resume_remoted_if_required()
